use std::fmt::Display;

use binrw::{BinRead, BinWrite};
use modular_bitfield::{
    bitfield,
    specifiers::{B1, B14},
};

use crate::{
    field_of::FieldOf,
    types::{Field, ObjectType, Point, WString, F64, U16, U32},
};

use self::{circle::Circle, ellipse::Ellipse, line::Lines, polygon::Polygon, rectangle::Rectangle};

pub mod circle;
pub mod ellipse;
pub mod hatch;
pub mod line;
pub mod polygon;
pub mod rectangle;

#[bitfield(bits = 16)]
#[derive(BinRead, BinWrite, Debug, Copy, Clone)]
#[br(map = Self::from_bytes)]
#[bw(map = |&x| Self::into_bytes(x))]
pub struct ObjectFlags {
    pub disabled: B1,
    pub aspect_ratio_unlocked: B1,
    #[skip]
    __: B14,
}

#[derive(BinRead, BinWrite, Debug)]
#[brw(magic(17u32))]
pub struct ObjectBase {
    pub pen: U32,
    pub obj_type: FieldOf<ObjectType>,
    pub flags: FieldOf<ObjectFlags>,
    pub name: WString,
    pub count: U32,
    _unknown_1: Field,
    pub io_control_enable_mask: U16,
    pub io_control_disable_mask: U16,
    _unknown_2: [Field; 5],
    pub origin: FieldOf<Point>,
    pub z: F64,
    pub a: F64,
    _unknown_3: Field,
}

impl Display for ObjectBase {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "[{} = disabled: {}, locked: {}, count: {}, origin: {}, z: {:.2}, a: {:.2}]",
            *self.obj_type,
            self.flags.disabled() != 0,
            self.flags.aspect_ratio_unlocked() != 0,
            *self.count,
            *self.origin,
            *self.z,
            *self.a,
        )
    }
}

#[derive(BinRead, BinWrite, Debug)]
pub enum Object {
    #[brw(magic = 1u32)]
    Curve(Lines),
    #[brw(magic = 2u32)]
    Point(Lines),
    #[brw(magic = 3u32)]
    Rectangle(Rectangle),
    #[brw(magic = 4u32)]
    Circle(Circle),
    #[brw(magic = 5u32)]
    Ellipse(Ellipse),
    #[brw(magic = 6u32)]
    Polygon(Polygon),
    // #[brw(magic = 32u32)]
    // Hatch(Hatch),
}

impl Display for Object {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Object::Curve(x) => write!(f, "{}", x),
            Object::Point(x) => write!(f, "{}", x),
            Object::Rectangle(x) => write!(f, "{}", x),
            Object::Circle(x) => write!(f, "{}", x),
            Object::Ellipse(x) => write!(f, "{}", x),
            Object::Polygon(x) => write!(f, "{}", x),
        }
    }
}
